/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2004 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@libsdl.org
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_x11gl_c.h,v 1.7.2.3 2004/02/23 04:53:54 slouken Exp $";
#endif

#ifdef HAVE_OPENGL
#include <GL/glx.h>
#include <dlfcn.h>
#if defined(__OpenBSD__) && !defined(__ELF__)
#define dlsym(x,y) dlsym(x, "_" y)
#endif
#endif
#include "SDL_sysvideo.h"

#ifndef GLX_SGIX_fbconfig
typedef struct __GLXFBConfigRec *GLXFBConfigSGIX;
#endif

#ifndef GLX_SGIX_pbuffer
typedef XID GLXPbufferSGIX;
#endif

struct SDL_PrivateGLData {
    int gl_active; /* to stop switching drivers while we have a valid context */
    int gl_extensions_checked;

#ifdef HAVE_OPENGL
#define MAKE_GLX_VERSION(major, minor)	((major << 16) + minor)
    int glx_version;
    GLXContext glx_context;	/* Current GL context */
    XVisualInfo* glx_visualinfo; /* XVisualInfo* returned by glXChooseVisual */

    void * (*glXGetProcAddress)(const GLubyte *procName);

    XVisualInfo* (*glXChooseVisual)
		( Display*		dpy,
		  int			screen,
		  int*			attribList );

    GLXContext (*glXCreateContext)
		( Display*		dpy,
		  XVisualInfo*		vis,
		  GLXContext		shareList,
		  Bool			direct );

    void (*glXDestroyContext)
		( Display* 		dpy,
		  GLXContext		ctx );

    Bool (*glXMakeCurrent)
		( Display*		dpy,
		  GLXDrawable		drawable,
		  GLXContext		ctx );

    void (*glXSwapBuffers)
		( Display*		dpy,
		  GLXDrawable		drawable );

    int (*glXGetConfig)
	 ( Display* dpy,
	   XVisualInfo* visual_info,
	   int attrib,
	   int* value );

    void (*glXReleaseBuffersMESA)
	 ( Display* dpy,
	   GLXDrawable drawable );

    const char *(*glXQueryExtensionsString)
	    ( Display* dpy,
	      int screen );

    /*
     * Support for GLX_SGIX_fbconfig and GLX_SGIX_pbuffer 
     */

    GLXFBConfigSGIX *(*glXChooseFBConfigSGIX)
	    ( Display* dpy,
	      int screen,
	      int *attrib_list,
	      int *nelements );

    int (*glXGetFBConfigAttribSGIX)
	    ( Display* dpy,
	      GLXFBConfigSGIX config,
	      int attribute,
	      int *value);

    GLXPbuffer (*glXCreateGLXPbufferSGIX)
	    ( Display* dpy,
	      GLXFBConfigSGIX config,
	      unsigned int width,
	      unsigned int height,
	      int *attrib_list );

    void (*glXDestroyGLXPbufferSGIX)
	    ( Display* dpy,
	      GLXPbuffer pbuf );
 
    /*
     * Support for GLX 1.3 pbuffer support
     */

    GLXPbufferSGIX (*glXCreatePbuffer)
	    ( Display* dpy,
	      GLXFBConfigSGIX config,
	      int *attrib_list );

#endif /* HAVE_OPENGL */
};

/* Old variable names */
#define gl_active		(this->gl_data->gl_active)
#define gl_extensions_checked	(this->gl_data->gl_extensions_checked)
#define glx_context		(this->gl_data->glx_context)
#define glx_visualinfo		(this->gl_data->glx_visualinfo)

/* OpenGL functions */
extern XVisualInfo *X11_GL_GetVisual(_THIS);
extern int X11_GL_CreateWindow(_THIS, int w, int h);
extern int X11_GL_CreateContext(_THIS);
extern void X11_GL_Shutdown(_THIS);
#ifdef HAVE_OPENGL
extern int X11_GL_MakeCurrent(_THIS);
extern int X11_GL_GetAttribute(_THIS, SDL_GLattr attrib, int* value);
extern void X11_GL_SwapBuffers(_THIS);
extern int X11_GL_LoadLibrary(_THIS, const char* path);
extern void *X11_GL_GetProcAddress(_THIS, const char* proc);
extern int X11_GL_CreateRenderTarget(_THIS, SDL_RenderTarget *target, int width, int height, int *attrib);
extern int X11_GL_GetRenderTargetAttribute(_THIS, SDL_RenderTarget *target, SDL_GLattr attrib, int* value);
extern void X11_GL_LockRenderTarget(_THIS, SDL_RenderTarget *target);
extern void X11_GL_UnlockRenderTarget(_THIS, SDL_RenderTarget *target);
extern void X11_GL_FreeRenderTarget(_THIS, SDL_RenderTarget *target);
#endif
extern void X11_GL_UnloadLibrary(_THIS);

#ifdef HAVE_OPENGL

#ifndef GLX_ARB_multisample
#define GLX_ARB_multisample
#define GLX_SAMPLE_BUFFERS_ARB             100000
#define GLX_SAMPLES_ARB                    100001
#endif

#ifndef GLX_SGIX_fbconfig
#define GLX_WINDOW_BIT_SGIX                0x00000001
#define GLX_PIXMAP_BIT_SGIX                0x00000002
#define GLX_RGBA_BIT_SGIX                  0x00000001
#define GLX_COLOR_INDEX_BIT_SGIX           0x00000002
#define GLX_DRAWABLE_TYPE_SGIX             0x8010
#define GLX_RENDER_TYPE_SGIX               0x8011
#define GLX_X_RENDERABLE_SGIX              0x8012
#define GLX_FBCONFIG_ID_SGIX               0x8013
#endif

#ifndef GLX_SGIX_pbuffer
#define GLX_PBUFFER_BIT_SGIX               0x00000004
#define GLX_PRESERVED_CONTENTS_SGIX        0x801B
#define GLX_LARGEST_PBUFFER_SGIX           0x801C
#endif

#endif /* HAVE_OPENGL */
