/*
 * NeHe Lesson 11
 * This code was created by Jeff Molofee '99 
 * (ported to Linux/SDL by Ti Leggett '01)
 * (modified to be a render target demo for SDL by Sam Lantinga '03)
 *
 * If you've found this code useful, please let me know.
 *
 * Visit Jeff at http://nehe.gamedev.net/
 * 
 * or for port-specific comments, questions, bugreports etc. 
 * email to leggett@eecs.tulane.edu
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "SDL.h"

#ifdef HAVE_OPENGL

#include "SDL_opengl.h"

/* screen width, height, and bit depth */
#define SCREEN_WIDTH  640
#define SCREEN_HEIGHT 480
#define SCREEN_BPP     16

/* render target width, height */
#define RENDERTARGET_WIDTH  256
#define RENDERTARGET_HEIGHT 256

/* Set up some booleans */
#define TRUE  1
#define FALSE 0

/* This is our SDL surface */
SDL_Surface *surface;

GLfloat xrot;            /* X Rotation */
GLfloat yrot;            /* Y Rotation */
GLfloat zrot;            /* Z Rotation */

float points[45][45][3]; /* The Points On The Grid Of Our "Wave" */
int wiggle_count = 0;    /* Counter Used To Control How Fast Flag Waves */
GLfloat hold;            /* Temporarily Holds A Floating Point Value */

GLuint texture[1];       /* Storage For One Texture */
GLfloat textureCoords[4];

SDL_RenderTarget *RenderTarget = NULL;
GLuint RenderTargetTextureID;

/* function to release/destroy our resources and restoring the old desktop */
void Quit( int returnCode )
{
    if ( RenderTarget ) {
        SDL_GL_FreeRenderTarget(RenderTarget);
    }
    glDeleteTextures(1, texture);
    glDeleteTextures( 1, &RenderTargetTextureID);

    /* clean up the window */
    SDL_Quit( );

    /* and exit appropriately */
    exit( returnCode );
}

/* function to load in bitmap as a GL texture */
static int power_of_two(int input)
{
	int value = 1;

	while ( value < input ) {
		value <<= 1;
	}
	return value;
}
GLuint SDL_GL_LoadTexture(SDL_Surface *surface, GLfloat *texcoord)
{
	GLuint texture;
	int w, h;
	SDL_Surface *image;
	SDL_Rect area;
	Uint32 saved_flags;
	Uint8  saved_alpha;

	/* Use the surface width and height expanded to powers of 2 */
	w = power_of_two(surface->w);
	h = power_of_two(surface->h);
	texcoord[0] = 0.0f;			/* Min X */
	texcoord[1] = 0.0f;			/* Min Y */
	texcoord[2] = (GLfloat)surface->w / w;	/* Max X */
	texcoord[3] = (GLfloat)surface->h / h;	/* Max Y */

	image = SDL_CreateRGBSurface(
			SDL_SWSURFACE,
			w, h,
			32,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN /* OpenGL RGBA masks */
			0x000000FF, 
			0x0000FF00, 
			0x00FF0000, 
			0xFF000000
#else
			0xFF000000,
			0x00FF0000, 
			0x0000FF00, 
			0x000000FF
#endif
		       );
	if ( image == NULL ) {
		return 0;
	}

	/* Save the alpha blending attributes */
	saved_flags = surface->flags&(SDL_SRCALPHA|SDL_RLEACCELOK);
	saved_alpha = surface->format->alpha;
	if ( (saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
		SDL_SetAlpha(surface, 0, 0);
	}

	/* Copy the surface into the GL texture image */
	area.x = 0;
	area.y = 0;
	area.w = surface->w;
	area.h = surface->h;
	SDL_BlitSurface(surface, &area, image, &area);

	/* Restore the alpha blending attributes */
	if ( (saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
		SDL_SetAlpha(surface, saved_flags, saved_alpha);
	}

	/* Create an OpenGL texture for the image */
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D,
		     0,
		     GL_RGBA,
		     w, h,
		     0,
		     GL_RGBA,
		     GL_UNSIGNED_BYTE,
		     image->pixels);
	SDL_FreeSurface(image); /* No longer needed */

	return texture;
}
int LoadGLTextures( )
{
    SDL_Surface *image = SDL_LoadBMP("sample.bmp");
    if ( image ) {
    	texture[0] = SDL_GL_LoadTexture(image, textureCoords);
	return TRUE;
    }
    return FALSE;
}

int CreateRenderTarget( )
{
    /* Create the texture that will be bound to the render target */
    glGenTextures(1, &RenderTargetTextureID);
    glBindTexture(GL_TEXTURE_2D, RenderTargetTextureID);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    /* Create the render target */
    RenderTarget = SDL_GL_CreateRenderTarget(RENDERTARGET_WIDTH, RENDERTARGET_HEIGHT, NULL);
    if ( RenderTarget ) {
	int value;
	printf("Created render target:\n");
	SDL_GL_GetRenderTargetAttribute( RenderTarget, SDL_GL_RED_SIZE, &value );
	printf( "SDL_GL_RED_SIZE: %d\n", value);
	SDL_GL_GetRenderTargetAttribute( RenderTarget, SDL_GL_GREEN_SIZE, &value );
	printf( "SDL_GL_GREEN_SIZE: %d\n", value);
	SDL_GL_GetRenderTargetAttribute( RenderTarget, SDL_GL_BLUE_SIZE, &value );
	printf( "SDL_GL_BLUE_SIZE: %d\n", value);
	SDL_GL_GetRenderTargetAttribute( RenderTarget, SDL_GL_ALPHA_SIZE, &value );
	printf( "SDL_GL_ALPHA_SIZE: %d\n", value);
	SDL_GL_GetRenderTargetAttribute( RenderTarget, SDL_GL_DEPTH_SIZE, &value );
	printf( "SDL_GL_DEPTH_SIZE: %d\n", value );

	SDL_GL_BindRenderTarget(RenderTarget, RenderTargetTextureID);
    } else {
        /* We can fake a render target in this demo by rendering to the
         * screen and copying to a texture before we do normal rendering.
         */
	printf("Failed to create render target, using screen buffer\n");

        glBindTexture(GL_TEXTURE_2D, RenderTargetTextureID);
        glTexImage2D(GL_TEXTURE_2D,
			0,
			GL_RGB,
			RENDERTARGET_WIDTH, RENDERTARGET_HEIGHT,
			0,
			GL_RGB,
			GL_UNSIGNED_BYTE,
			NULL);
    }

    return TRUE;
}

/* to resize the window and reset our view */
int resizeWindow( int width, int height )
{
    /* Height / width ration */
    GLfloat ratio;

    /* Protect against a divide by zero */
    if ( height == 0 )
	height = 1;

    ratio = ( GLfloat )width / ( GLfloat )height;

    /* Setup our viewport. */
    glViewport( 0, 0, ( GLint )width, ( GLint )height );

    /* change to the projection matrix and set our viewing volume. */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );

    /* Set our perspective */
    gluPerspective( 45.0f, ratio, 0.1f, 100.0f );

    /* Make sure we're chaning the model view and not the projection */
    glMatrixMode( GL_MODELVIEW );

    /* Reset The View */
    glLoadIdentity( );

    return( TRUE );
}

/* function to handle key press events */
void handleKeyPress( SDL_keysym *keysym )
{
    switch ( keysym->sym )
	{
	case SDLK_ESCAPE:
	    /* ESC key was pressed */
	    Quit( 0 );
	    break;
	case SDLK_F1:
	    /* F1 key was pressed
	     * this toggles fullscreen mode
	     */
	    SDL_WM_ToggleFullScreen( surface );
	    break;
	default:
	    break;
	}

    return;
}

/* general OpenGL initialization function */
int initGL( GLvoid )
{
    int x, y; /* Loop variables */

    /* Load in the texture */
    if ( !LoadGLTextures( ) )
	return FALSE;

    if ( !CreateRenderTarget( ) )
	return FALSE;

    /* Enable Texture Mapping ( NEW ) */
    glEnable( GL_TEXTURE_2D );

    /* Enable smooth shading */
    glShadeModel( GL_SMOOTH );

    /* Set the background black */
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );

    /* Depth buffer setup */
    glClearDepth( 1.0f );

    /* Enables Depth Testing */
    glEnable( GL_DEPTH_TEST );

    /* The Type Of Depth Test To Do */
    glDepthFunc( GL_LEQUAL );

    /* Really Nice Perspective Calculations */
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

    /* Fill the back with the texture. The front will only be wireline */
    glPolygonMode( GL_FRONT, GL_LINE );
    glPolygonMode( GL_BACK, GL_FILL );

    /* Loop Through The X Plane */
    for ( x = 0; x < 45; x++ )
        {
	    /* Loop Through The Y Plane */
	    for ( y = 0; y < 45; y++ )
                {
		    /* Apply The Wave To Our Mesh */
		    points[x][y][0] = ( float )( ( x / 5.0f ) - 4.5f );
		    points[x][y][1] = ( float )( ( y / 5.0f ) - 4.5f );
		    points[x][y][2] = ( float )( sin( ( ( ( x / 5.0f ) * 40.0f ) / 360.0f ) * 3.141592654 * 2.0f ) );
                }
        }

    return( TRUE );
}

int drawRenderTarget( GLvoid )
{
    static int initialized = 0;
    static GLfloat cubeModelView[16];
    static float color[8][3]= {{ 1.0,  1.0,  0.0}, 
                               { 1.0,  0.0,  0.0},
                               { 0.0,  0.0,  0.0},
                               { 0.0,  1.0,  0.0},
                               { 0.0,  1.0,  1.0},
                               { 1.0,  1.0,  1.0},
                               { 1.0,  0.0,  1.0},
                               { 0.0,  0.0,  1.0}};
    static float cube[8][3]= {{ 0.5,  0.5, -0.5}, 
                              { 0.5, -0.5, -0.5},
                              {-0.5, -0.5, -0.5},
                              {-0.5,  0.5, -0.5},
                              {-0.5,  0.5,  0.5},
                              { 0.5,  0.5,  0.5},
                              { 0.5, -0.5,  0.5},
                              {-0.5, -0.5,  0.5}};

    /* Set initial state */
    if ( RenderTarget ) {
        SDL_GL_LockRenderTarget(RenderTarget);
    }
    glPushAttrib( GL_ALL_ATTRIB_BITS ); /* Overkill, but safe */

    if ( RenderTarget ) {
    	glViewport( 0, 0, RenderTarget->w, RenderTarget->h );
    } else {
    	glViewport( 0, 0, RENDERTARGET_WIDTH, RENDERTARGET_HEIGHT );
    }

    glMatrixMode( GL_PROJECTION );
    glPushMatrix();

    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();

    glDisable( GL_FOG );
    glDisable( GL_BLEND );
    glDisable( GL_ALPHA_TEST );
    glDisable( GL_SCISSOR_TEST );	
    glDisable( GL_STENCIL_TEST );
    glDisable( GL_CULL_FACE );

    glShadeModel( GL_SMOOTH );
    glPolygonMode( GL_FRONT, GL_FILL );
    glPolygonMode( GL_BACK, GL_FILL );

    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );

    /* Clear the background before we start */
    glClearColor( 0.0, 0.0, 0.0, 1.0 );
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    /* Render the logo texture to the background */
    glEnable( GL_TEXTURE_2D );
    glDisable( GL_DEPTH_TEST );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
    glOrtho( 0.0, 1.0, 0.0, 1.0, 0.0, 1.0 );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity( );

    glBindTexture( GL_TEXTURE_2D, texture[0] );
    glBegin( GL_TRIANGLE_STRIP );
        glTexCoord2f(textureCoords[0], textureCoords[3]);
	glVertex2f(0.0, 0.0);
        glTexCoord2f(textureCoords[2], textureCoords[3]);
	glVertex2f(1.0, 0.0);
        glTexCoord2f(textureCoords[0], textureCoords[1]);
	glVertex2f(0.0, 1.0);
        glTexCoord2f(textureCoords[2], textureCoords[1]);
	glVertex2f(1.0, 1.0);
    glEnd();

    /* Render the spinning cube */
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc( GL_LESS );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
    glOrtho( -2.0, 2.0, -2.0, 2.0, -20.0, 20.0 );

    glMatrixMode( GL_MODELVIEW );
    if ( initialized ) {
        glLoadMatrixf(cubeModelView);
    	glRotatef(5.0, 1.0, 1.0, 1.0);
    } else {
        glLoadIdentity();
	initialized = 1;
    }
    glGetFloatv( GL_MODELVIEW_MATRIX, cubeModelView);

    glBegin( GL_QUADS );

        glColor3fv(color[0]);
        glVertex3fv(cube[0]);
        glColor3fv(color[1]);
        glVertex3fv(cube[1]);
        glColor3fv(color[2]);
        glVertex3fv(cube[2]);
        glColor3fv(color[3]);
        glVertex3fv(cube[3]);
        
        glColor3fv(color[3]);
        glVertex3fv(cube[3]);
        glColor3fv(color[4]);
        glVertex3fv(cube[4]);
        glColor3fv(color[7]);
        glVertex3fv(cube[7]);
        glColor3fv(color[2]);
        glVertex3fv(cube[2]);
        
        glColor3fv(color[0]);
        glVertex3fv(cube[0]);
        glColor3fv(color[5]);
        glVertex3fv(cube[5]);
        glColor3fv(color[6]);
        glVertex3fv(cube[6]);
        glColor3fv(color[1]);
        glVertex3fv(cube[1]);
        
        glColor3fv(color[5]);
        glVertex3fv(cube[5]);
        glColor3fv(color[4]);
        glVertex3fv(cube[4]);
        glColor3fv(color[7]);
        glVertex3fv(cube[7]);
        glColor3fv(color[6]);
        glVertex3fv(cube[6]);

        glColor3fv(color[5]);
        glVertex3fv(cube[5]);
        glColor3fv(color[0]);
        glVertex3fv(cube[0]);
        glColor3fv(color[3]);
        glVertex3fv(cube[3]);
        glColor3fv(color[4]);
        glVertex3fv(cube[4]);

        glColor3fv(color[6]);
        glVertex3fv(cube[6]);
        glColor3fv(color[1]);
        glVertex3fv(cube[1]);
        glColor3fv(color[2]);
        glVertex3fv(cube[2]);
        glColor3fv(color[7]);
        glVertex3fv(cube[7]);

    glEnd( );
    
    /* Clean up */
    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();

    glMatrixMode( GL_PROJECTION );
    glPopMatrix();

    glPopAttrib();

    if ( RenderTarget ) {
	/* Note that we could also use the glCopyTexSubImage2D trick below
	 * before unlocking the render target, but some drivers have an
	 * optimized path if you bind the texture beforehand.  You could
	 * still use it if you wanted to get the depth or stencil buffers.
	 */
        SDL_GL_UnlockRenderTarget(RenderTarget);
    } else {
        /* Copy our rendering to the fake render target texture */
        glBindTexture( GL_TEXTURE_2D, RenderTargetTextureID );
        glCopyTexSubImage2D( GL_TEXTURE_2D,
                          0,
		          0, 0,
		          0, 0, RENDERTARGET_WIDTH, RENDERTARGET_HEIGHT);
    }

    {
        /* Check for error conditions. */
        GLenum gl_error = glGetError( );

        if( gl_error != GL_NO_ERROR ) {
            fprintf( stderr, "testrendertarget: opengl error: %d (0x%4.4x)\n", gl_error, gl_error );
        }
    }

    return( TRUE );
}

/* Here goes our drawing code */
int drawGLScene( GLvoid )
{
    /* These are to calculate our fps */
    static GLint T0     = 0;
    static GLint Frames = 0;

    int x, y;                   /* Loop Variables */
    float f_x, f_y, f_xb, f_yb; /* Used To Break The Flag Into Tiny Quads */
 
    /* Render into our render target texture */
    drawRenderTarget();

    /* Clear The Screen And Depth Buffer */
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    /* Reset The Current Matrix */
    glLoadIdentity( );

    /* Translate 17 Units Into The Screen */
    glTranslatef( 0.0f, 0.0f, -12.0f );

    glRotatef( xrot, 1.0f, 0.0f, 0.0f ); /* Rotate On The X Axis */
    glRotatef( yrot, 0.0f, 1.0f, 0.0f ); /* Rotate On The Y Axis */
    glRotatef( zrot, 0.0f, 0.0f, 1.0f ); /* Rotate On The Z Axis */

    glBindTexture( GL_TEXTURE_2D, RenderTargetTextureID ); /* Select Our Texture */

    /* Start Drawing Our Quads */
    glBegin( GL_QUADS );
    /* Loop Through The X Plane 0-44 (45 Points) */
    for( x = 0; x < 44; x++ )
	{
	    /* Loop Through The Y Plane 0-44 (45 Points) */
	    for( y = 0; y < 44; y++ )
		{
		    /* Create A Floating Point X Value */
		    f_x = ( float )x / 44.0f;
		    /* Create A Floating Point Y Value */
		    f_y = ( float )y / 44.0f;
		    /* Create A Floating Point Y Value+0.0227f */
		    f_xb = ( float )( x + 1 ) / 44.0f;
		    /* Create A Floating Point Y Value+0.0227f */
		    f_yb = ( float )( y + 1 ) / 44.0f;

		    /* First Texture Coordinate (Bottom Left) */
		    glTexCoord2f( f_x, f_y );
		    glVertex3f( points[x][y][0], points[x][y][1],
				points[x][y][2] );

		    /* Second Texture Coordinate (Top Left) */
		    glTexCoord2f( f_x, f_yb );
		    glVertex3f( points[x][y + 1][0], points[x][y + 1][1],
				points[x][y + 1][2] );

		    /* Third Texture Coordinate (Top Right) */
		    glTexCoord2f( f_xb, f_yb );
		    glVertex3f( points[x + 1][y + 1][0],
				points[x + 1][y + 1][1],
				points[x + 1][y + 1][2] );

		    /* Fourth Texture Coordinate (Bottom Right) */
		    glTexCoord2f( f_xb, f_y );
		    glVertex3f( points[x + 1][y][0], points[x + 1][y][1],
				points[x + 1][y][2] );
                }
        }
    glEnd( );

    /* Used To Slow Down The Wave (Every 2nd Frame Only) */
    if( wiggle_count == 2 )
        {
	    /* Loop Through The Y Plane */
	    for( y = 0; y < 45; y++ )
		{
		    /* Store Current Value One Left Side Of Wave */
		    hold = points[0][y][2];
		    /* Loop Through The X Plane */
		    for( x = 0; x < 44; x++)
			{
                            /* Current Wave Value Equals Value To The Right */
			    points[x][y][2] = points[x + 1][y][2];
                        }
		    /* Last Value Becomes The Far Left Stored Value */
		    points[44][y][2] = hold;
                }
	    wiggle_count = 0; /* Set Counter Back To Zero */
        }
    wiggle_count++; /* Increase The Counter */

    xrot += 0.3f; /* Increase The X Rotation Variable */
    yrot += 0.2f; /* Increase The Y Rotation Variable */
    zrot += 0.4f; /* Increase The Z Rotation Variable */

    /* Draw it to the screen */
    SDL_GL_SwapBuffers( );

    /* Gather our frames per second */
    Frames++;
    {
	GLint t = SDL_GetTicks();
	if (t - T0 >= 5000) {
	    GLfloat seconds = (t - T0) / 1000.0f;
	    GLfloat fps = Frames / seconds;
	    printf("%d frames in %g seconds = %g FPS\n", Frames, seconds, fps);
	    T0 = t;
	    Frames = 0;
	}
    }

    return( TRUE );
}

int main( int argc, char **argv )
{
    /* Flags to pass to SDL_SetVideoMode */
    int videoFlags;
    /* main loop variable */
    int done = FALSE;
    /* used to collect events */
    SDL_Event event;
    /* whether or not the window is active */
    int isActive = TRUE;

    /* initialize SDL */
    if ( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE ) < 0 )
	{
	    fprintf( stderr, "Video initialization failed: %s\n",
		     SDL_GetError( ) );
	    Quit( 1 );
	}

    /* the flags to pass to SDL_SetVideoMode */
    videoFlags  = SDL_OPENGL;          /* Enable OpenGL in SDL */
    videoFlags |= SDL_RESIZABLE;       /* Enable window resizing */

    /* Sets up OpenGL double buffering */
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /* get a SDL surface */
    surface = SDL_SetVideoMode( SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP,
				videoFlags );

    /* Verify there is a surface */
    if ( !surface )
	{
	    fprintf( stderr,  "Video mode set failed: %s\n", SDL_GetError( ) );
	    Quit( 1 );
	}

    /* initialize OpenGL */
    if ( !initGL( ) )
	{
	    fprintf( stderr,  "Couldn't initialize textures\n");
	    Quit( 1 );
	}

    /* resize the initial window */
    resizeWindow( SCREEN_WIDTH, SCREEN_HEIGHT );

    /* wait for events */
    while ( !done )
	{
	    /* handle the events in the queue */

	    while ( SDL_PollEvent( &event ) )
		{
		    switch( event.type )
			{
			case SDL_ACTIVEEVENT:
			    /* Something's happend with our focus
			     * If we lost focus or we are iconified, we
			     * shouldn't draw the screen
			     */
			    if ( event.active.state & SDL_APPACTIVE ) {
			        if ( event.active.gain == 0 )
				    isActive = FALSE;
			        else
				    isActive = TRUE;
			    }
			    break;			    
			case SDL_VIDEORESIZE:
			    /* handle resize event */
			    surface = SDL_SetVideoMode( event.resize.w,
							event.resize.h,
							16, videoFlags );
			    if ( !surface )
				{
				    fprintf( stderr, "Could not get a surface after resize: %s\n", SDL_GetError( ) );
				    Quit( 1 );
				}
			    resizeWindow( event.resize.w, event.resize.h );
			    break;
			case SDL_KEYDOWN:
			    /* handle key presses */
			    handleKeyPress( &event.key.keysym );
			    break;
			case SDL_QUIT:
			    /* handle quit requests */
			    done = TRUE;
			    break;
			default:
			    break;
			}
		}

	    /* draw the scene */
	    if ( isActive )
		drawGLScene( );
	}

    /* clean ourselves up and exit */
    Quit( 0 );

    /* Should never get here */
    return( 0 );
}

#else /* HAVE_OPENGL */

int main(int argc, char *argv[])
{
	printf("No OpenGL support on this system\n");
	return 1;
}

#endif /* HAVE_OPENGL */
